# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for creating a named selection."""

from typing import TYPE_CHECKING

from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import grpc_point_to_point3d
from ansys.geometry.core.designer.beam import Beam
from ansys.geometry.core.designer.body import Body
from ansys.geometry.core.designer.component import Component
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.designer.edge import Edge
from ansys.geometry.core.designer.face import Face
from ansys.geometry.core.designer.vertex import Vertex
from ansys.geometry.core.misc.auxiliary import (
    get_beams_from_ids,
    get_bodies_from_ids,
    get_components_from_ids,
    get_edges_from_ids,
    get_faces_from_ids,
    get_vertices_from_ids,
)

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.design import Design


class NamedSelection:
    """Represents a single named selection within the design assembly.

    This class synchronizes to a design within a supporting Geometry service instance.

    A named selection organizes one or more design entities together for common actions
    against the entire group.

    Parameters
    ----------
    name : str
        User-defined name for the named selection.
    design : Design
        Design instance to which the named selection belongs.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    bodies : list[Body], default: None
        All bodies to include in the named selection.
    faces : list[Face], default: None
        All faces to include in the named selection.
    edges : list[Edge], default: None
        All edges to include in the named selection.
    beams : list[Beam], default: None
        All beams to include in the named selection.
    design_points : list[DesignPoints], default: None
        All design points to include in the named selection.
    components: list[Component], default: None
        All components to include in the named selection.
    """

    def __init__(
        self,
        name: str,
        design: "Design",
        grpc_client: GrpcClient,
        bodies: list[Body] | None = None,
        faces: list[Face] | None = None,
        edges: list[Edge] | None = None,
        beams: list[Beam] | None = None,
        design_points: list[DesignPoint] | None = None,
        components: list[Component] | None = None,
        vertices: list[Vertex] | None = None,
        preexisting_id: str | None = None,
    ):
        """Initialize the ``NamedSelection`` class."""
        self._name = name
        self._design = design
        self._grpc_client = grpc_client

        # Create empty arrays if there are none of a type
        if bodies is None:
            bodies = []
        if faces is None:
            faces = []
        if edges is None:
            edges = []
        if beams is None:
            beams = []
        if design_points is None:
            design_points = []
        if components is None:
            components = []
        if vertices is None:
            vertices = []

        # Instantiate
        self._bodies = bodies
        self._faces = faces
        self._edges = edges
        self._beams = beams
        self._design_points = design_points
        self._components = components
        self._vertices = vertices

        # Store ids for later use... when verifying if the NS changed.
        self._ids_cached = {
            "bodies": [body.id for body in bodies],
            "faces": [face.id for face in faces],
            "edges": [edge.id for edge in edges],
            "beams": [beam.id for beam in beams],
            "design_points": [dp.id for dp in design_points],
            "components": [component.id for component in components],
            "vertices": [vertex.id for vertex in vertices],
        }

        if preexisting_id:
            self._id = preexisting_id
            return

        # All ids should be unique - no duplicated values
        ids = set()

        # Loop over all entities to get their ids
        for value in self._ids_cached.values():
            for entity_id in value:
                ids.add(entity_id)

        response = self._grpc_client.services.named_selection.create_named_selection(
            name=name, members=ids
        )
        self._id = response.get("id")

    @property
    def id(self) -> str:
        """ID of the named selection."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the named selection."""
        return self._name

    @property
    def bodies(self) -> list[Body]:
        """All bodies in the named selection."""
        self.__verify_ns()
        if self._bodies is None:
            # Get all bodies from the named selection
            self._bodies = get_bodies_from_ids(self._design, self._ids_cached["bodies"])

        return self._bodies

    @property
    def faces(self) -> list[Face]:
        """All faces in the named selection."""
        self.__verify_ns()
        if self._faces is None:
            # Get all faces from the named selection
            self._faces = get_faces_from_ids(self._design, self._ids_cached["faces"])

        return self._faces

    @property
    def edges(self) -> list[Edge]:
        """All edges in the named selection."""
        self.__verify_ns()
        if self._edges is None:
            # Get all edges from the named selection
            self._edges = get_edges_from_ids(self._design, self._ids_cached["edges"])

        return self._edges

    @property
    def beams(self) -> list[Beam]:
        """All beams in the named selection."""
        self.__verify_ns()
        if self._beams is None:
            # Get all beams from the named selection
            self._beams = get_beams_from_ids(self._design, self._ids_cached["beams"])

        return self._beams

    @property
    def design_points(self) -> list[DesignPoint]:
        """All design points in the named selection."""
        self.__verify_ns()
        if self._design_points is None:
            # Get all design points from the named selection
            self._design_points = [
                DesignPoint(dp_id, f"dp: {dp_id}", grpc_point_to_point3d(dp_point))
                for dp_id, dp_point in self._ids_cached["design_points"]
            ]

        return self._design_points

    @property
    def components(self) -> list[Component]:
        """All components in the named selection."""
        self.__verify_ns()
        if self._grpc_client.backend_version < (26, 1, 0):
            self._grpc_client.log.warning(
                "Accessing components in named selections is only"
                " consistent starting in version 2026 R1."
            )
            return []
        if self._components is None:
            # Get all components from the named selection
            self._components = get_components_from_ids(self._design, self._ids_cached["components"])

        return self._components

    @property
    def vertices(self) -> list[Vertex]:
        """All vertices in the named selection."""
        self.__verify_ns()
        if self._grpc_client.backend_version < (26, 1, 0):
            self._grpc_client.log.warning(
                "Accessing vertices of named selections is only"
                " consistent starting in version 2026 R1."
            )
            return []
        if self._vertices is None:
            # Get all vertices from the named selection
            self._vertices = get_vertices_from_ids(self._design, self._ids_cached["vertices"])

        return self._vertices

    def __verify_ns(self) -> None:
        """Verify that the contents of the named selection are up to date."""
        if self._grpc_client.backend_version < (25, 2, 0):
            self._grpc_client.log.warning(
                "Accessing members of named selections is only"
                " consistent starting in version 2025 R2."
            )
            return

        # Get all entities from the named selection
        response = self._grpc_client.services.named_selection.get_named_selection(id=self._id)

        # Check if the named selection has changed
        ids = {
            "bodies": response.get("bodies"),
            "faces": response.get("faces"),
            "edges": response.get("edges"),
            "beams": response.get("beams"),
            "design_points": response.get("design_points"),
            "components": response.get("components"),
            "vertices": response.get("vertices"),
        }

        for key in ids:
            if ids[key] != self._ids_cached[key]:
                # Clear the cache for that specific entity
                setattr(self, f"_{key}", None)
                # Update the cache
                self._ids_cached[key] = ids[key]

    def __repr__(self) -> str:
        """Represent the ``NamedSelection`` as a string."""
        lines = [f"ansys.geometry.core.designer.selection.NamedSelection {hex(id(self))}"]
        lines.append(f"  Name                 : {self._name}")
        lines.append(f"  Id                   : {self._id}")
        lines.append(f"  N Bodies             : {len(self.bodies)}")
        lines.append(f"  N Faces              : {len(self.faces)}")
        lines.append(f"  N Edges              : {len(self.edges)}")
        lines.append(f"  N Beams              : {len(self.beams)}")
        lines.append(f"  N Design Points      : {len(self.design_points)}")
        lines.append(f"  N Components         : {len(self.components)}")
        lines.append(f"  N Vertices           : {len(self.vertices)}")
        return "\n".join(lines)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for creating and managing gears."""

from typing import TYPE_CHECKING

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity

from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.checks import graphics_required
from ansys.geometry.core.misc.measurements import Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.arc import Arc
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.segment import SketchSegment
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv


class Gear(SketchFace):
    """Provides the base class for sketching gears."""

    def __init__(self):
        """Initialize the ``Gear`` class."""
        super().__init__()

    @property
    @graphics_required
    def visualization_polydata(self) -> "pv.PolyData":
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import pyvista as pv

        return pv.merge([edge.visualization_polydata for edge in self._edges])


class DummyGear(Gear):
    """Provides the dummy class for sketching gears.

    Parameters
    ----------
    origin : Point2D
        Origin of the gear.
    outer_radius : ~pint.Quantity | Distance | Real
        Outer radius of the gear.
    inner_radius : ~pint.Quantity | Distance | Real
        Inner radius of the gear.
    n_teeth : int
        Number of teeth of the gear.
    """

    @check_input_types
    def __init__(
        self,
        origin: Point2D,
        outer_radius: Quantity | Distance | Real,
        inner_radius: Quantity | Distance | Real,
        n_teeth: int,
    ):
        """Initialize the ``DummyGear`` class."""
        # Call the parent ctor
        super().__init__()

        # Ensure radiuses are Distances
        outer_radius = (
            outer_radius if isinstance(outer_radius, Distance) else Distance(outer_radius)
        )
        inner_radius = (
            inner_radius if isinstance(inner_radius, Distance) else Distance(inner_radius)
        )

        # Compute auxiliary variables
        repeat_angle = 2 * np.pi / n_teeth
        tooth_angle = 0.475 * repeat_angle
        gap_angle = repeat_angle - tooth_angle

        # Now, loop over all teeth to build the sketch
        for tooth_idx in range(n_teeth):
            # Three angles need to be computed: starting tooth angle,
            # ending tooth angle (==starting gap angle), and ending gap angle
            start_ang = tooth_idx * repeat_angle
            inter_ang = start_ang + tooth_angle
            end_ang = inter_ang + gap_angle

            # Compute the sin and cos values for these angles
            sin_inter = np.sin(inter_ang)
            cos_inter = np.cos(inter_ang)
            sin_end = np.sin(end_ang)
            cos_end = np.cos(end_ang)

            # Define the points for drawing the arcs and segments involved
            outer_arc_start = Point2D(
                [
                    outer_radius.value.m * np.cos(start_ang),
                    outer_radius.value.m * np.sin(start_ang),
                ],
                unit=outer_radius.unit,
            )
            outer_arc_end = Point2D(
                [outer_radius.value.m * cos_inter, outer_radius.value.m * sin_inter],
                unit=outer_radius.unit,
            )
            inner_arc_start = Point2D(
                [inner_radius.value.m * cos_inter, inner_radius.value.m * sin_inter],
                unit=inner_radius.unit,
            )
            inner_arc_end = Point2D(
                [inner_radius.value.m * cos_end, inner_radius.value.m * sin_end],
                unit=inner_radius.unit,
            )
            next_outer_arc_start = Point2D(
                [outer_radius.value.m * cos_end, outer_radius.value.m * sin_end],
                unit=outer_radius.unit,
            )

            # Now, proceed to draw the arcs and segments
            # TODO: add plane to SketchSegment when available
            # https://github.com/ansys/pyansys-geometry/issues/1319
            self._edges.append(
                Arc(start=outer_arc_start + origin, end=outer_arc_end + origin, center=origin)
            )
            self._edges.append(
                SketchSegment(start=outer_arc_end + origin, end=inner_arc_start + origin)
            )
            self._edges.append(
                Arc(start=inner_arc_start + origin, end=inner_arc_end + origin, center=origin)
            )
            self._edges.append(
                SketchSegment(start=inner_arc_end + origin, end=next_outer_arc_start + origin)
            )


class SpurGear(Gear):
    """Provides the class for sketching spur gears.

    Parameters
    ----------
    origin : Point2D
        Origin of the spur gear.
    module : Real
        Module of the spur gear. This is also the ratio between the pitch circle
        diameter in millimeters and the number of teeth.
    pressure_angle : ~pint.Quantity | Angle | Real
        Pressure angle of the spur gear.
    n_teeth : int
        Number of teeth of the spur gear.
    """

    @check_input_types
    def __init__(
        self,
        origin: Point2D,
        module: Real,
        pressure_angle: Quantity | Angle | Real,
        n_teeth: int,
    ):
        """Initialize spur gears."""
        # Call the parent ctor
        super().__init__()

        # Additional checks for inputs
        pressure_angle = (
            pressure_angle if isinstance(pressure_angle, Angle) else Angle(pressure_angle)
        )

        # Store input parameters
        self._origin = origin
        self._module = module
        self._pressure_angle = pressure_angle.value.to(UNITS.radian)
        self._n_teeth = n_teeth

        # Compute additional needed values
        self._ref_diameter = self.module * self.n_teeth
        self._base_diameter = self.ref_diameter * np.cos(self.pressure_angle.m)
        self._addendum = self.module
        self._dedendum = 1.25 * self.module
        self._tip_diameter = self.ref_diameter + 2 * self.module
        self._root_diameter = self.ref_diameter - 2.5 * self.module

        # Sketch the gear
        self._sketch_spur_gear()

    @property
    def origin(self) -> Point2D:
        """Origin of the spur gear."""
        return self._origin

    @property
    def module(self) -> Real:
        """Module of the spur gear."""
        return self._module

    @property
    def pressure_angle(self) -> Quantity:
        """Pressure angle of the spur gear."""
        return self._pressure_angle

    @property
    def n_teeth(self) -> int:
        """Number of teeth of the spur gear."""
        return self._n_teeth

    @property
    def ref_diameter(self) -> Real:
        """Reference diameter of the spur gear."""
        return self._ref_diameter

    @property
    def base_diameter(self) -> Real:
        """Base diameter of the spur gear."""
        return self._base_diameter

    @property
    def addendum(self) -> Real:
        """Addendum of the spur gear."""
        return self._addendum

    @property
    def dedendum(self) -> Real:
        """Dedendum of the spur gear."""
        return self._dedendum

    @property
    def tip_diameter(self) -> Real:
        """Tip diameter of the spur gear."""
        return self._tip_diameter

    @property
    def root_diameter(self) -> Real:
        """Root diameter of the spur gear."""
        return self._root_diameter

    def _sketch_spur_gear(self) -> None:
        """Generate the spur gear sketch from the properties defined."""
        # Sketch a single tooth first
        tooth_lines = self._sketch_single_tooth_spur_gear()

        # Now, for all teeth, rotate those values
        rotate_angle = 2 * np.pi / self.n_teeth
        last_point = None
        for tooth_idx in range(self.n_teeth):
            # Rotate the tooth points by a given angle
            x_i, y_i = self._rotate_curve(tooth_idx * rotate_angle, tooth_lines[0], tooth_lines[2])
            x_m, y_m = self._rotate_curve(tooth_idx * rotate_angle, tooth_lines[1], tooth_lines[3])

            if last_point:
                # Add the closing arc from the previous tooth
                self._edges.extend(
                    self._generate_arcs(
                        [last_point[0], x_i[0]], [last_point[1], y_i[0]], closing_involute=True
                    )
                )

            # Generate the arcs from involute curve
            self._edges.extend(self._generate_arcs(x_i, y_i))
            # Add the closing involute-to-mirror arc
            self._edges.extend(
                self._generate_arcs([x_i[-1], x_m[0]], [y_i[-1], y_m[0]], closing_involute=True)
            )
            # Generate the arcs from mirrored involute curve
            self._edges.extend(self._generate_arcs(x_m, y_m))
            # Update the last point value
            last_point = (x_m[-1], y_m[-1])

        # When coming out of the loop, close with the starting tooth
        self._edges.extend(
            self._generate_arcs(
                [last_point[0], tooth_lines[0][0]],
                [last_point[1], tooth_lines[2][0]],
                closing_involute=True,
            )
        )

    def _sketch_single_tooth_spur_gear(
        self,
    ) -> tuple[list[Real], list[Real], list[Real], list[Real]]:
        """Sketch a single tooth using this private method.

        Returns
        -------
        tuple[list[Real], list[Real], list[Real], list[Real]]
            X and Y values for the tooth grouped together as follows: (x_i, x_m, y_i, y_m)
        """
        # Compute the involute with the smallest of both
        #
        # FYI: Max angle is slightly less than  90deg
        diam = self.root_diameter if self.root_diameter < self.base_diameter else self.base_diameter
        x_i, y_i, t_i = self._involute(diam / 2, self.tip_diameter / 2, np.pi / 2.1)

        # Align the involute curve to cross the pitch diameter
        x_i, y_i = self._align_involute(x_i, y_i, t_i)

        # Get the mirrored section of the involute
        x_m = list(reversed(x_i))
        y_m = [-y for y in list(reversed(y_i))]

        # Rotate the mirrored curve by the circular tooth angle
        # First, compute the tooth thickness = circular pitch (== module * pi) / (2 + backlash)
        # Then, the angle will be = thickness * 2 / diametral pitch
        circular_tooth_thickness = (self.module * np.pi) / 2.05
        circular_tooth_angle = circular_tooth_thickness * 2 / (self.module * self.n_teeth)
        x_m, y_m = self._rotate_curve(circular_tooth_angle, x_m, y_m)

        # Now that you have the whole tooth points, return them
        return (x_i, x_m, y_i, y_m)

    def _involute(
        self, radius: Real, max_radius: Real, max_theta: Real, steps: int = 30
    ) -> tuple[list[Real], list[Real]]:
        """Generate the involute points discretization of a curve.

        Parameters
        ----------
        radius : Real
            Departing radius for computing the involute.
        max_radius : Real
            Maximum radius up to which to compute the involute.
        max_theta : Real
            Maximum angle up to which to compute the involute.
        steps : int, default: 30
            Number of steps to use to discretize the curve.

        Returns
        -------
        tuple[list[Real], list[Real], list[Real]]
            Three-element tuple containing a list of X, Y, and theta values
            defining the involute.
        """
        # Instantiate the containers for storing the results
        x_p = []
        y_p = []
        t_p = []

        # Define the delta angle to be increased in each step
        dtheta = max_theta / steps

        # Iterate over the defined steps
        for i in range(steps):
            # Compute the involute X, Y and curve angle values
            c_dtheta = np.cos(i * dtheta)
            s_dtheta = np.sin(i * dtheta)
            invol_x = radius * (c_dtheta + i * dtheta * s_dtheta)
            invol_y = radius * (s_dtheta - i * dtheta * c_dtheta)
            invol_ang = np.arctan2(invol_y, invol_x)
            dist = np.sqrt(invol_x**2 + invol_y**2)

            # Append values to result containers
            x_p.append(invol_x)
            y_p.append(invol_y)
            t_p.append(invol_ang)

            # Check if you overcame the max_radius...
            if dist > max_radius:
                # You passed the limit... readjust (linear interp to the max)
                adjustment = (max_radius - radius) / (dist - radius)
                x_p[-1] = x_p[-2] * (1 - adjustment) + invol_x * adjustment
                y_p[-1] = y_p[-2] * (1 - adjustment) + invol_y * adjustment
                t_p[-1] = t_p[-2] * (1 - adjustment) + invol_ang * adjustment
                break

        return (x_p, y_p, t_p)

    def _align_involute(
        self, x_p: list[Real], y_p: list[Real], t_p: list[Real]
    ) -> tuple[list[Real], list[Real]]:
        """Align the discretized values of the involute curve.

        Parameters
        ----------
        x_p : list[Real]
            X-elements defining the involute curve to align.
        y_p : list[Real]
            Y-elements defining the involute curve to align.
        t_p : list[Real]
            Angles defining the involute curve to align.

        Returns
        -------
        tuple[list[Real], list[Real]]
            Set of X and Y elements that have been aligned.

        Raises
        ------
        ValueError
            If no alignment angle is found.
        """
        # Compute the angle where the involute curve crosses the circle
        theta_cross = None
        pitch_circle_radius = (self.module * self.n_teeth) / 2
        rr = pitch_circle_radius * pitch_circle_radius
        for idx in range(0, len(x_p) - 1):
            rr_2 = x_p[idx + 1] * x_p[idx + 1] + y_p[idx + 1] * y_p[idx + 1]
            if rr_2 > rr:
                # This means we passed the angle, adjust!
                r1 = np.sqrt(x_p[idx] * x_p[idx] + y_p[idx] * y_p[idx])
                r2 = np.sqrt(rr_2)
                adjustment = (pitch_circle_radius - r1) / (r2 - r1)
                theta_cross = t_p[idx] * (1 - adjustment) + t_p[idx + 1] * adjustment
                break

        # If no angle is found... fail!
        if not theta_cross:  # pragma: no cover
            raise ValueError("Error in involute alignment. Check values and implementation.")

        # Proceed to alignment using -theta_cross
        return self._rotate_curve(-theta_cross, x_p, y_p)

    def _rotate_curve(
        self, angle: Real, x_p: list[Real], y_p: list[Real]
    ) -> tuple[list[Real], list[Real]]:
        """Rotate X,Y elements defining a curve by a given angle.

        Parameters
        ----------
        angle : Real
            Angle (in radians) to rotate the X,Y elements.
        x_p : list[Real]
            X-elements of the curve to rotate.
        y_p : list[Real]
            Y-elements of the curve to rotate.

        Returns
        -------
        tuple[list[Real], list[Real]]
            The X and Y elements of the rotated curve.
        """
        # Compute the sin and cos values of the angle
        c_ang = np.cos(angle)
        s_ang = np.sin(angle)

        # Rotate the points
        x_r = []
        y_r = []
        for x, y in zip(x_p, y_p):
            x_r.append(c_ang * x - s_ang * y)
            y_r.append(s_ang * x + c_ang * y)

        return (x_r, y_r)

    def _generate_arcs(
        self, x_p: list[Real], y_p: list[Real], closing_involute: bool = False
    ) -> list[Arc]:
        """Generate the arcs of involute curves when sketching spur gears.

        Parameters
        ----------
        x_p : list[Real]
            X-elements defining the involute curve.
        y_p : list[Real]
            Y-elements defining the involute curve.
        closing_involute : bool, optional
            Shortcut for joining involute curves, by default False.

        Returns
        -------
        list[Arc]
            The list of arcs defining the requested curve.
        """
        # Initialize results container
        arcs = []

        # Generate Point2D objects from given X, Y values (remember, they are in mm)
        points = [
            Point2D(
                [x_i + self.origin.x.to(UNITS.mm).m, y_i + self.origin.y.to(UNITS.mm).m],
                unit=UNITS.mm,
            )
            for (x_i, y_i) in zip(x_p, y_p)
        ]

        if not closing_involute:
            # Initiate preliminary arc object - lives outside the scope of the loop
            preliminary_arc = None

            for idx in range(len(points) - 2):
                # Compute a preliminary arc taking into account three points
                preliminary_arc = Arc.from_three_points(
                    start=points[idx], inter=points[idx + 1], end=points[idx + 2]
                )

                # Keep only the first two points as part of the arc... use the
                # needed values from the preliminary arc
                arcs.append(
                    Arc(
                        start=points[idx],
                        end=points[idx + 1],
                        center=preliminary_arc.center,
                        clockwise=preliminary_arc.is_clockwise,
                    )
                )

            # Once the loop has finished... extend the last arc
            arcs.append(
                Arc(
                    start=points[-2],
                    end=points[-1],
                    center=preliminary_arc.center,
                    clockwise=preliminary_arc.is_clockwise,
                )
            )

        else:
            # We should only enter this branch if we are closing the involute curves...
            # ... which means that we only have two points in our list...
            #
            # Just in case, let us only take the first and last elements of the given list
            arcs.append(
                Arc(
                    start=points[0],
                    end=points[-1],
                    center=self.origin,
                )
            )

        # Finally, return the requested arcs
        return arcs



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the designs service implementation for v0."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.designs import GRPCDesignsService
from .conversions import (
    _check_write_body_facets_input,
    build_grpc_id,
    from_design_file_format_to_grpc_part_export_format,
)


class GRPCDesignsServiceV0(GRPCDesignsService):  # pragma: no cover
    """Designs service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    designs service. It is specifically designed for the v0 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.dbu.v0.designs_pb2_grpc import DesignsStub

        self.stub = DesignsStub(channel)

    @protect_grpc
    def open(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.designs_pb2 import OpenRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = OpenRequest(
            filepath=kwargs["filepath"],
            import_options=kwargs["import_options"].to_dict(),
        )

        # Call the gRPC service
        _ = self.stub.Open(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def new(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.designs_pb2 import NewRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = NewRequest(name=kwargs["name"])

        # Call the gRPC service
        response = self.stub.New(request)

        # Return the response - formatted as a dictionary
        return {
            "design_id": response.id,
            "main_part_id": response.main_part.id,
        }

    @protect_grpc
    def close(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(id=kwargs["design_id"])

        # Call the gRPC service
        _ = self.stub.Close(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def put_active(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(id=kwargs["design_id"])

        # Call the gRPC service
        _ = self.stub.PutActive(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def save_as(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.designs_pb2 import SaveAsRequest

        _check_write_body_facets_input(kwargs["backend_version"], kwargs["write_body_facets"])

        # Create the request - assumes all inputs are valid and of the proper type
        request = SaveAsRequest(
            filepath=kwargs["filepath"], write_body_facets=kwargs["write_body_facets"]
        )

        # Call the gRPC service
        _ = self.stub.SaveAs(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def download_export(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.designs_pb2 import DownloadExportFileRequest

        _check_write_body_facets_input(kwargs["backend_version"], kwargs["write_body_facets"])

        # Create the request - assumes all inputs are valid and of the proper type
        request = DownloadExportFileRequest(
            format=from_design_file_format_to_grpc_part_export_format(kwargs["format"]),
            write_body_facets=kwargs["write_body_facets"],
        )

        # Call the gRPC service
        response = self.stub.DownloadExportFile(request)

        # Return the response - formatted as a dictionary
        data = bytes()
        data += response.data
        return {"data": data}

    @protect_grpc
    def stream_download_export(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.designs_pb2 import DownloadExportFileRequest

        _check_write_body_facets_input(kwargs["backend_version"], kwargs["write_body_facets"])

        # Create the request - assumes all inputs are valid and of the proper type
        request = DownloadExportFileRequest(
            format=from_design_file_format_to_grpc_part_export_format(kwargs["format"]),
            write_body_facets=kwargs["write_body_facets"],
        )

        # Call the gRPC service
        response = self.stub.StreamDownloadExportFile(request)

        # Return the response - formatted as a dictionary
        data = bytes()
        for elem in response:
            data += elem.data

        return {"data": data}

    @protect_grpc
    def insert(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.designs_pb2 import InsertRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = InsertRequest(
            filepath=kwargs["filepath"], import_named_selections=kwargs["import_named_selections"]
        )

        # Call the gRPC service
        _ = self.stub.Insert(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def get_active(self, **kwargs) -> dict:  # noqa: D102
        from google.protobuf.empty_pb2 import Empty

        # Call the gRPC service
        response = self.stub.GetActive(request=Empty())

        # Return the response - formatted as a dictionary
        if response:
            return {
                "design_id": response.id,
                "main_part_id": response.main_part.id,
                "name": response.name,
            }



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""PyAnsys Geometry connection subpackage."""

from ansys.geometry.core.connection.backend import ApiVersions, BackendType
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import (
    curve_to_grpc_curve,
    frame_to_grpc_frame,
    grpc_curve_to_curve,
    grpc_frame_to_frame,
    grpc_matrix_to_matrix,
    grpc_surface_to_surface,
    line_to_grpc_line,
    plane_to_grpc_plane,
    point3d_to_grpc_point,
    sketch_shapes_to_grpc_geometries,
    trimmed_curve_to_grpc_trimmed_curve,
    unit_vector_to_grpc_direction,
)
import ansys.geometry.core.connection.defaults as defaults
from ansys.geometry.core.connection.docker_instance import (
    GeometryContainers,
    LocalDockerInstance,
    get_geometry_container_type,
)
from ansys.geometry.core.connection.launcher import (
    launch_docker_modeler,
    launch_modeler,
    launch_modeler_with_core_service,
    launch_modeler_with_discovery,
    launch_modeler_with_discovery_and_pimlight,
    launch_modeler_with_geometry_service,
    launch_modeler_with_geometry_service_and_pimlight,
    launch_modeler_with_spaceclaim,
    launch_modeler_with_spaceclaim_and_pimlight,
    launch_remote_modeler,
)
from ansys.geometry.core.connection.product_instance import ProductInstance



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the faces service implementation for v0."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.conversions import to_area, to_distance
from ..base.faces import GRPCFacesService
from .conversions import (
    build_grpc_id,
    from_grpc_curve_to_curve,
    from_grpc_point_to_point3d,
    from_grpc_surface_to_surface,
)


class GRPCFacesServiceV0(GRPCFacesService):  # pragma: no cover
    """Faces service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    faces service. It is specifically designed for the v0 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2_grpc import FacesStub

        self.stub = FacesStub(channel)

    @protect_grpc
    def get_surface(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetSurface(request=request)

        # Return the response - formatted as a dictionary
        return {
            "surface": from_grpc_surface_to_surface(response, kwargs["surface_type"]),
        }

    @protect_grpc
    def get_box_uv(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetBoxUV(request=request)

        # Return the response - formatted as a dictionary
        return {
            "uv_box": {
                "u": (response.start_u, response.end_u),
                "v": (response.start_v, response.end_v),
            }
        }

    @protect_grpc
    def get_area(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetArea(request=request)

        # Return the response - formatted as a dictionary
        return {"area": to_area(response.area)}

    @protect_grpc
    def get_edges(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetEdges(request=request)

        # Return the response - formatted as a dictionary
        return {
            "edges": [
                {
                    "id": edge.id,
                    "curve_type": edge.curve_type,
                    "is_reversed": edge.is_reversed,
                }
                for edge in response.edges
            ]
        }

    @protect_grpc
    def get_vertices(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        from ansys.api.geometry.v0.faces_pb2 import GetVerticesRequest

        request = GetVerticesRequest(face_id=build_grpc_id(kwargs["id"]))

        # Call the gRPC service
        response = self.stub.GetVertices(request=request)

        # Return the response - formatted as a dictionary
        return {
            "vertices": [
                {
                    "id": vertex.id.id,
                    "position": from_grpc_point_to_point3d(vertex.position),
                }
                for vertex in response.vertices
            ]
        }

    @protect_grpc
    def get_loops(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetLoops(request=request)

        # Return the response - formatted as a dictionary
        return {
            "loops": [
                {
                    "type": loop.type,
                    "length": to_distance(loop.length).value,
                    "min_corner": from_grpc_point_to_point3d(loop.bounding_box.min),
                    "max_corner": from_grpc_point_to_point3d(loop.bounding_box.max),
                    "edges": [edge for edge in loop.edges],
                }
                for loop in response.loops
            ]
        }

    @protect_grpc
    def get_color(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetColor(request=request)

        # Return the response - formatted as a dictionary
        return {"color": response.color}

    @protect_grpc
    def get_bounding_box(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetBoundingBox(request=request)

        # Return the response - formatted as a dictionary
        return {
            "min_corner": from_grpc_point_to_point3d(response.min),
            "max_corner": from_grpc_point_to_point3d(response.max),
            "center": from_grpc_point_to_point3d(response.center),
        }

    @protect_grpc
    def set_color(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import SetColorRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetColorRequest(
            face_id=kwargs["id"],
            color=kwargs["color"],
        )
        # Call the gRPC service
        response = self.stub.SetColor(request=request)

        # Return the response - formatted as a dictionary
        return {"success": response.success}

    @protect_grpc
    def get_normal(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import GetNormalRequest
        from ansys.geometry.core.math.vector import UnitVector3D

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetNormalRequest(
            id=kwargs["id"],
            u=kwargs["u"],
            v=kwargs["v"],
        )

        # Call the gRPC service
        response = self.stub.GetNormal(request=request)

        # Return the response - formatted as a dictionary
        return {
            "normal": UnitVector3D(
                [response.direction.x, response.direction.y, response.direction.z]
            ),
        }

    @protect_grpc
    def evaluate(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import EvaluateRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = EvaluateRequest(
            id=kwargs["id"],
            u=kwargs["u"],
            v=kwargs["v"],
        )

        # Call the gRPC service
        response = self.stub.Evaluate(request=request)

        # Return the response - formatted as a dictionary
        return {
            "point": from_grpc_point_to_point3d(response.point),
        }

    @protect_grpc
    def create_iso_parametric_curve(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import CreateIsoParamCurvesRequest
        from ansys.geometry.core.shapes.parameterization import Interval

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateIsoParamCurvesRequest(
            id=kwargs["id"],
            u_dir_curve=kwargs["use_u_param"],
            proportion=kwargs["parameter"],
        )

        # Call the gRPC service
        response = self.stub.CreateIsoParamCurves(request=request)

        # Return the response - formatted as a dictionary
        return {
            "curves": [
                {
                    "geometry": from_grpc_curve_to_curve(curve.curve),
                    "start": from_grpc_point_to_point3d(curve.start),
                    "end": from_grpc_point_to_point3d(curve.end),
                    "interval": Interval(curve.interval_start, curve.interval_end),
                    "length": to_distance(curve.length).value,
                }
                for curve in response.curves
            ]
        }



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for repair tool message."""

from typing import TYPE_CHECKING

from ansys.api.geometry.v0.repairtools_pb2 import RepairGeometryRequest
from ansys.api.geometry.v0.repairtools_pb2_grpc import RepairToolsStub
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.tools.repair_tool_message import RepairToolMessage

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body


class GeometryIssue:
    """Provides return message for the repair tool methods."""

    def __init__(
        self,
        message_type: str,
        message_id: str,
        message: str,
        edges: list[str],
        faces: list[str],
    ):
        """Initialize a new instance of a geometry issue found during geometry inspect.

        Parameters
        ----------
        message_type: str
            Type of the message (warning, error, info).
        message_id: str
            Identifier for the message.
        message
            Message that describes the geometry issue.
        edges: list[str]
            List of edges (if any) that are part of the issue.
        modified_bodies: list[str]
            List of faces that are part of the issue.
        """
        self._message_type = message_type
        self._message_id = message_id
        self._message = message
        self._edges = edges
        self._faces = faces

    @property
    def message_type(self) -> str:
        """The type of the message (warning, error, info)."""
        return self._message_type

    @property
    def message_id(self) -> str:
        """The identifier for the message."""
        return self._message_id

    @property
    def message(self) -> str:
        """The content of the message."""
        return self._message

    @property
    def edges(self) -> list[str]:
        """The List of edges (if any) that are part of the issue."""
        return self._edges

    @property
    def faces(self) -> list[str]:
        """The List of faces (if any) that are part of the issue."""
        return self._faces


class InspectResult:
    """Provides the result of the inspect geometry operation."""

    def __init__(self, grpc_client: GrpcClient, body: "Body", issues: list[GeometryIssue]):
        """Initialize a new instance of the result of the inspect geometry operation.

        Parameters
        ----------
        body: Body
            Body for which issues are found.
        issues: list[GeometryIssue]
            List of issues for the body.
        """
        self._body = body
        self._issues = issues
        self._repair_stub = RepairToolsStub(grpc_client.channel)

    @property
    def body(self) -> "Body":
        """The body for which issues are found."""
        return self._body

    @property
    def issues(self) -> list[GeometryIssue]:
        """The list of issues for the body."""
        return self._issues

    def repair(self) -> RepairToolMessage:
        """Repair the problem area.

        Returns
        -------
        RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.body:
            return RepairToolMessage(False, [], [])

        repair_result_response = self._repair_stub.RepairGeometry(
            RepairGeometryRequest(bodies=[self.body._grpc_id])
        )

        return RepairToolMessage(repair_result_response.result.success, [], [])
